/*
 * Copyright 2021 VIA Technologies, Inc. All Rights Reserved.
 *
 * This PROPRIETARY SOFTWARE is the property of VIA Technologies, Inc.
 * and may contain trade secrets and/or other confidential information of
 * VIA Technologies, Inc. This file shall not be disclosed to any third
 * party, in whole or in part, without prior written consent of VIA.
 *
 * THIS PROPRIETARY SOFTWARE AND ANY RELATED DOCUMENTATION ARE PROVIDED AS IS,
 * WITH ALL FAULTS, AND WITHOUT WARRANTY OF ANY KIND EITHER EXPRESS OR IMPLIED,
 * AND VIA TECHNOLOGIES, INC. DISCLAIMS ALL EXPRESS OR IMPLIED
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, QUIET
 * ENJOYMENT OR NON-INFRINGEMENT.
*/
#include <Harpcam.h>
#include <esp_camera.h>
#include <base64.h>

Harpcam::Harpcam()
{
	uuid_ = "";
	funcID_ = 0;
	outputID_ = 0;
	numObj_ = 0;
	detected_ = false;
}

bool Harpcam::setupCamera()
{
	Serial.println("setupCamera");
	camera_config_t config;
	config.ledc_channel = LEDC_CHANNEL_0;
	config.ledc_timer = LEDC_TIMER_0;
	config.pin_d0 = Y2_GPIO_NUM;
	config.pin_d1 = Y3_GPIO_NUM;
	config.pin_d2 = Y4_GPIO_NUM;
	config.pin_d3 = Y5_GPIO_NUM;
	config.pin_d4 = Y6_GPIO_NUM;
	config.pin_d5 = Y7_GPIO_NUM;
	config.pin_d6 = Y8_GPIO_NUM;
	config.pin_d7 = Y9_GPIO_NUM;
	config.pin_xclk = XCLK_GPIO_NUM;
	config.pin_pclk = PCLK_GPIO_NUM;
	config.pin_vsync = VSYNC_GPIO_NUM;
	config.pin_href = HREF_GPIO_NUM;
	config.pin_sscb_sda = SIOD_GPIO_NUM;
	config.pin_sscb_scl = SIOC_GPIO_NUM;
	config.pin_pwdn = PWDN_GPIO_NUM;
	config.pin_reset = RESET_GPIO_NUM;
	config.xclk_freq_hz = 20000000;
	config.pixel_format = PIXFORMAT_JPEG;

	config.frame_size = FRAMESIZE_QVGA;
  	config.jpeg_quality = 10;
  	config.fb_count = 2;

	// camera init
	esp_err_t err = esp_camera_init(&config);
	if (err != ESP_OK) {
		Serial.printf("Camera init failed with error 0x%x", err);
		return false;
	}

	sensor_t *s = esp_camera_sensor_get();
	s->set_hmirror(s, 0);
	s->set_vflip(s, 0);

	return true;
}

String Harpcam::begin()
{
	StaticJsonDocument<128> doc;
	doc["header"] = "VERSION";
	doc["id"] = 99;
	if (uuid_.length() > 0)
		doc["uuid"] = uuid_.c_str();

	String msg;
	serializeJson(doc, msg);

	return msg;
}

String Harpcam::process()
{
	String cmd;

	switch (state_) {
		case Harpcam::CMD:
			cmd = packFuncCMD();
			break;
		case Harpcam::IMG:
			cmd = packImgCMD();
			break;
	}

	return cmd;
}

bool Harpcam::parseJsonContent(String msg_data)
{
	DynamicJsonDocument json(1024);

	detected_ = false;
	numObj_ = 0;

	auto deserializeError = deserializeJson(json, msg_data);
	if (deserializeError) {
		Serial.println("JSON parseObject() failed");
		return false;
	}

	String header = json["header"];

	if (header == "VERSION") {
		String ver = json["firmware"];
		Serial.printf("version: %s\n", ver);
		state_ = Harpcam::CMD;
	}
	else if (header == "SUCCESS") {
		int id = json["id"];
		Serial.printf("id: %d\n", id);
		state_ = Harpcam::IMG;
	}
	else if (header == "DETECT" || header == "KEYPOINT") {
		parseDetectionResult(json);
		state_ = Harpcam::IMG;
	}

	return true;
}

void Harpcam::parseDetectionResult(DynamicJsonDocument json)
{
	int id = json["id"];
	int num = json["num"];

	outputID_ = id;

	if (num > 0) {
		if (num > MAX_OBJS && id != Harpcam::FUNC_LANES_DETECTION)
			num = MAX_OBJS;
		detected_ = true;
		numObj_ = num;
		memset(&outputData_, 0, sizeof(OutputData)*MAX_OBJS);
		if (id == Harpcam::FUNC_APRILTAG)
			memset(&apriltagData_, 0, sizeof(ApriltagExtraData)*MAX_OBJS);
		else if (id == Harpcam::FUNC_LANES_DETECTION)
			memset(&lanesData_, 0, sizeof(LanesExtraData));
	}

	if (id != Harpcam::FUNC_LANES_DETECTION) {
		for (int i=0; i<num; i++) {
			/*Serial.printf("i:%d, type:%d, x:%d, y:%d, w:%d, h:%d c:%f\n", 
				i, outputData_[i].type, outputData_[i].x, outputData_[i].y, outputData_[i].w, outputData_[i].h,
				outputData_[i].confidence);*/
			outputData_[i].type = json["objects"][i]["type"];
			outputData_[i].x = json["objects"][i]["x"];
			outputData_[i].y = json["objects"][i]["y"];
			outputData_[i].w = json["objects"][i]["w"];
			outputData_[i].h = json["objects"][i]["h"];
			outputData_[i].confidence = json["objects"][i]["confidence"];

			if (id == Harpcam::FUNC_APRILTAG) {
				apriltagData_[i].posx = json["objects"][i]["posx"];
				apriltagData_[i].posy = json["objects"][i]["posy"];
				apriltagData_[i].posz = json["objects"][i]["posz"];
				apriltagData_[i].rotx = json["objects"][i]["rotx"];
				apriltagData_[i].roty = json["objects"][i]["roty"];
				apriltagData_[i].centerx = json["objects"][i]["cx"];
				apriltagData_[i].centery = json["objects"][i]["cy"];
			}
		}
	}
	else {
		if (num == 5) {
			lanesData_.centerx = json["objects"][0]["x"];
			lanesData_.centery = json["objects"][0]["y"];
			lanesData_.lx1 = json["objects"][1]["x"];
			lanesData_.ly1 = json["objects"][1]["y"];
			lanesData_.lx2 = json["objects"][2]["x"];
			lanesData_.ly2 = json["objects"][2]["y"];
			lanesData_.rx1 = json["objects"][3]["x"];
			lanesData_.ry1 = json["objects"][3]["y"];
			lanesData_.rx2 = json["objects"][4]["x"];
			lanesData_.ry2 = json["objects"][4]["y"];
		}
	}
}

bool Harpcam::isDetected()
{
	return detected_;
}

void Harpcam::setUUID(String uuid)
{
	uuid_ = uuid;
}

void Harpcam::setFuncID(int id)
{
	funcID_ = id;
}

void Harpcam::getOutputData(OutputData *data)
{
	if (numObj_ > 0)
		memcpy(data, &outputData_, sizeof(OutputData)*numObj_);
}

void Harpcam::getApriltagExtraData(ApriltagExtraData *data)
{
	if (numObj_ > 0)
		memcpy(data, &apriltagData_, sizeof(ApriltagExtraData)*numObj_);
}

void Harpcam::getLanesExtraData(LanesExtraData *data)
{
	if (numObj_ > 0)
		memcpy(data, &lanesData_, sizeof(LanesExtraData));
}

int Harpcam::getNumOfObjects()
{
	return numObj_;
}

int Harpcam::getOutputFuncID()
{
	return outputID_;
}

int Harpcam::getFuncID()
{
	return funcID_;
}

String Harpcam::packFuncCMD()
{
	StaticJsonDocument<256> doc;
	doc["header"] = "FUNC";
	doc["id"] = funcID_;
	doc["realtime"] = true;

	String cmd;
	serializeJson(doc, cmd);

	Serial.printf("packFuncCMD: ");
	Serial.println(cmd);

	return cmd;
}

String Harpcam::packImgCMD()
{
	String cmd;
	camera_fb_t *fb = NULL;

	fb = esp_camera_fb_get();
	if (!fb) {
		Serial.println("Camera capture failed");
		return cmd;
	}

	String base64str = base64::encode((uint8_t *)fb->buf, fb->len);

	DynamicJsonDocument doc(1024);
	doc["header"] = "IMAGE";
	doc["img"] = base64str.c_str();

	serializeJson(doc, cmd);

	esp_camera_fb_return(fb);

	return cmd;
}